# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: November 21st 2020

Modified By: hsky77
Last Updated: March 3rd 2021 16:53:19 pm
'''


import os

from ..util import join_path, join_to_abs_path, BaseLocal
from .component import ComponentManager, ComponentTypes, create_component_manager
from .constants import Component_Module_Folder, LocalCode_File_Not_Found, Project_Config_File


class ProjectMixin():
    @property
    def project_dir(self) -> str:
        return self._project_dir

    @property
    def project_config_path(self) -> str:
        return join_to_abs_path(self.project_dir, Project_Config_File)

    @property
    def project_config(self):
        return self._project_config

    @property
    def component_manager(self) -> ComponentManager:
        return self._component_manager

    @property
    def name(self) -> str:
        return self.project_config['name']

    @property
    def debug(self) -> bool:
        return self.project_config['debug']

    def load_project(self, project_dir: str) -> None:
        self._project_dir = join_to_abs_path(project_dir)
        self._project_config = {}
        self._component_manager = None

        with open(self.project_config_path, 'r', encoding='utf8') as f:
            import yaml
            from .config_validator import ProjectConfigValidator
            validator = ProjectConfigValidator(
                yaml.load(f, Loader=yaml.SafeLoader))
            self._project_config = validator.parameter

        self._component_manager = create_component_manager(
            self.project_dir, self.project_config[Component_Module_Folder]) if Component_Module_Folder in self.project_config else create_component_manager(self.project_dir)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: October 18th 2021 09:03:51 am
'''

import os
import logging

from ...util import join_path, BaseSyncLogger, LOG_FORMAT


class FileLoggerMixin():
    def update_file_handler(self, logger: BaseSyncLogger, sub_dir: str = '', mode: str = 'a', encoding: str = 'utf-8', **kwargs):
        """
        Remove logger's file handler or update it with the specfied 'log_dir'.
        """
        log_dir = kwargs.get('dir', None)

        if log_dir:
            log_dir = join_path(kwargs.get(
                'project_dir', ''), log_dir, sub_dir)
            if not os.path.isdir(log_dir):
                os.makedirs(log_dir)

            log_file = join_path(log_dir, logger.name + '.log')

            exist = False
            for h in logger.handlers:
                if type(h) is logging.FileHandler:
                    if h.baseFilename == os.path.abspath(log_file):
                        exist = True
                    else:
                        h.close()
                        logger.removeHandler(h)

            if not exist:
                handler = logging.FileHandler(
                    log_file, mode=mode, encoding=encoding)
                handler.setFormatter(logging.Formatter(LOG_FORMAT))
                logger.addHandler(handler)

    def remove_file_handler(self, logger: BaseSyncLogger, sub_dir: str = '', **kwargs):
        log_dir = kwargs.get('dir', None)

        if log_dir:
            log_file = join_path(kwargs.get('project_dir', ''),
                                 log_dir, sub_dir, logger.name + '.log')

            for h in logger.handlers:
                if type(h) is logging.FileHandler and h.baseFilename == os.path.abspath(log_file):
                    h.close()
                    logger.removeHandler(h)
                    break



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: September 3rd 2020 14:06:01 pm
'''

from .constants import *
from .utils import *
from .dynamic_class_enum import DynamicTypeEnum
from .hierarchy_element import HierarchyElementMeta
from .localization import *
from .logger import *
from .worker import *
from .executor import *
from .callbacks import Callbacks

BaseLocal.import_csv([join_path(__path__[0], Localization_File)])



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

    - class Executor runs the given functions synchronously or asynchronously
    - class ExecutorFactory manages FunctionQueueWorker instances and produces the Executors.

Modified By: hsky77
Last Updated: September 3rd 2020 14:30:27 pm
'''


import time
import random
import asyncio
from typing import Any, Callable, List

from .worker import FunctionQueueWorker


class _WorkerTask():
    def __init__(self):
        self.result = None
        self.exception = None
        self.done = False

    def on_finish(self, result: Any):
        self.result = result
        self.done = True

    def on_exception(self, e: Exception):
        self.exception = e
        self.done = True


class Executor():
    def __init__(self, worker: FunctionQueueWorker, *args, **kwargs):
        self.__worker = worker

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, type, value, traceback):
        pass

    @property
    def worker(self):
        return self.__worker

    def run_method_in_queue(self,
                            func: Callable,
                            *args,
                            on_finish: Callable[[Any], None] = None,
                            on_exception: Callable[[Exception], None] = None,
                            **kwargs) -> None:
        """Run the given functions. It does not block the calling thread."""
        self.__worker.run_method(
            func, *args, on_finish=on_finish, on_exception=on_exception, **kwargs)

    def run_method(self,
                   func: Callable,
                   *args,
                   **kwargs) -> Any:
        """Run the given function. It blocks the calling thread."""
        task = _WorkerTask()
        self.__worker.run_method(
            func, *args, on_finish=task.on_finish, on_exception=task.on_exception, **kwargs)

        while not task.done:
            time.sleep(0)

        if task.exception:
            raise task.exception from task.exception
        return task.result

    async def run_method_async(self,
                               func: Callable,
                               *args,
                               **kwargs) -> Any:
        """Run the given function asynchronously."""
        task = _WorkerTask()
        self.__worker.run_method(
            func, *args, on_finish=task.on_finish, on_exception=task.on_exception, **kwargs)

        while not task.done:
            await asyncio.sleep(0)

        if task.exception:
            raise task.exception from task.exception
        return task.result


class ExecutorFactory():
    def __init__(self,
                 pool_name: str = None,
                 executor_type: Executor = Executor,
                 worker_type: FunctionQueueWorker = FunctionQueueWorker,
                 worker_limit: int = 1):
        self._pool_name = pool_name or type(self).__name__
        self._executor_type = executor_type
        self._worker_type = worker_type
        self._worker_limit = worker_limit
        self._pool = []
        self._disposed = False

    @property
    def worker_count(self) -> int:
        return len(self._pool)

    @property
    def workers(self) -> List[FunctionQueueWorker]:
        return self._pool

    def dispose(self):
        if not self._disposed:
            self._disposed = True
            for w in self._pool:
                w.dispose()

    def run_method_in_queue(self,
                            func: Callable,
                            *args,
                            on_finish: Callable[[Any], None] = None,
                            on_exception: Callable[[Exception], None] = None,
                            **kwargs) -> None:
        """Run the given func in queue. It does not block the calling thread"""
        with self.get_executor() as executor:
            executor.run_method_in_queue(
                func, *args, on_finish=on_finish, on_exception=on_exception, **kwargs)

    def run_method(self,
                   func: Callable,
                   *args,
                   **kwargs) -> Any:
        """
        Run the given func. It blocks the calling thread.
        """
        with self.get_executor() as executor:
            return executor.run_method(func, *args, **kwargs)

    async def run_method_async(self,
                               func: Callable,
                               *args,
                               **kwargs) -> Any:
        """Run the given func asynchronously."""
        async with self.get_executor() as executor:
            return await executor.run_method_async(func, *args, **kwargs)

    def get_executor(self, *args, **kwargs) -> Executor:
        """
        Create and return an executor instance.
        """
        if len(self._pool) < self._worker_limit:
            worker = self._worker_type(
                '{}_{}'.format(self._pool_name, len(self._pool)))
            self._pool.append(worker)
            return self._executor_type(worker, *args, **kwargs)

        return self._executor_type(self._pool[random.randint(0, len(self._pool)-1)], *args, **kwargs)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: December 27th 2020 06:40:42 am
'''


from enum import Enum, EnumMeta
from typing import Type, List

from .localization import BaseLocal
from .constants import LocalCode_Must_Be_3Str_Tuple, LocalCode_Must_Be_Class, LocalCode_Must_Be_Function, LocalCode_Is_Not_Subclass


class DynamicTypeEnumMeta(EnumMeta):
    """Customized enum meta class to import class dynamically"""
    def __new__(metacls, cls, bases, classdict):
        cls = super().__new__(metacls, cls, bases, classdict)
        cls._check_element()
        return cls

    def __call__(cls: Enum, value, names=None, *, module=None, qualname=None, type=None, start=1):
        if names is None:
            return cls._create_instance(value)
        return cls._create_(value, names, module=module, qualname=qualname, type=type, start=start)

    def _create_instance(cls: Enum, value):
        for member in cls.__members__.values():
            if member.value[0] == value:
                return member

        return cls.__new__(cls, value)

    def _check_element(cls: Enum):
        for e in cls:
            if type(e.value) is not tuple and type(e.value) is not list:
                raise TypeError(BaseLocal.get_message(
                    LocalCode_Must_Be_3Str_Tuple))
            else:
                if len(e.value) == 3:
                    for s in e.value:
                        if not isinstance(s, str):
                            raise TypeError(BaseLocal.get_message(
                                LocalCode_Must_Be_3Str_Tuple))
                else:
                    raise TypeError(BaseLocal.get_message(
                        LocalCode_Must_Be_3Str_Tuple))


class DynamicTypeEnum(Enum, metaclass=DynamicTypeEnumMeta):
    """
    Customized enum class to import class dynamically.

    Every element must be 3 string tuple as ('enum_type_key', 'module', 'class or function')

    example:
        class TestType(DynamicTypeEnum):
            Class1 = ('class1', 'module', 'class or function')
    """

    def import_class(self, cls_type: type = None) -> Type:
        from . import get_class
        from inspect import isclass
        try:
            cls = get_class(self._get_module_name(),
                            self.enum_object_type_name)
        except:
            cls = get_class(self.enum_module_name, self.enum_object_type_name)

        if not isclass(cls):
            raise TypeError(BaseLocal.get_message(
                LocalCode_Must_Be_Class, cls))

        if cls_type is not None and not issubclass(cls, cls_type):
            raise TypeError(BaseLocal.get_message(
                LocalCode_Is_Not_Subclass, cls, cls_type))

        return cls

    def import_function(self) -> Type:
        from . import get_class
        from inspect import isfunction
        try:
            func = get_class(self._get_module_name(),
                             self.enum_object_type_name)
        except:
            func = get_class(self.enum_module_name, self.enum_object_type_name)

        if not isfunction(func):
            raise TypeError(BaseLocal.get_message(
                LocalCode_Must_Be_Function, func))
        return func

    def _get_module_name(self):
        module_name = self.__module__
        if self.enum_module_name is not None or not self.enum_module_name == '':
            module_name = '{}.{}'.format(module_name, self.enum_module_name)
        return module_name

    @property
    def enum_key(self) -> str:
        return self.value[0]

    @property
    def enum_module_name(self) -> str:
        return self.value[1]

    @property
    def enum_object_type_name(self) -> str:
        return self.value[2]

    @staticmethod
    def get_dynamic_class_enum_class(module_dir: str) -> List[Enum]:
        from importlib import import_module
        from inspect import isclass
        m = import_module(module_dir)
        return [cls for cls in [getattr(m, k) for k in dir(m)] if isclass(
            cls) and issubclass(cls, DynamicTypeEnum)]



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

provide basic python unittest.TestSuite of the following test cases:
    - hyssop.util, worker, and executor
    - hyssop.project.web config validation, components and controllers

use get_test_suite() to get the default or extend test suite by inheriting "hyssop.unit_test.UnitTestTypes."

Usage:
    - to create unit_test extension module:

        0. the hierarchy of folders and files looks like:

            server_directory/
                server_config.yaml
                unit_test/
                    __init__.py
                    foo.py

        1. "foo.py" defines class FooTestCase:

            from hyssop.unit_test import UnitTestCase
            
            class FooTestCase(UnitTestCase):
                # override this test method
                def test(self):
                    # test modules...

        2. "__init__.py" defines the enum class allows load the test case classes dynamically,

            from hyssop.unit_test import UnitTestTypes

            class ExTestTypes(UnitTestTypes):
                FooTest = ('foo_test', 'foo', 'FooTestCase')

        3. In the commond prompt, run command "python -m hyssop test <server_directory>" 
            to test all the extend test cases defined in "__init__.py"

Modified By: hsky77
Last Updated: November 22nd 2020 12:23:32 pm
'''

import unittest

from .base import UnitTestTypes, UnitTestCase


class DefaultUnitTestTypes(UnitTestTypes):
    TestUtil = ('test_util', 'util', 'UtilTestCase')
    TestWorker = ('test_worker', 'util_worker', 'WorkerTestCase')


def get_test_suite() -> unittest.TestSuite:
    """get test suite of unittest module. 
    It will try to load extend test suite if specifed in server folder "unit_test", elsewise default test suite.
    Default test suite tests util and web modules of hyssop
    """

    suite = unittest.TestSuite()
    # extension tests
    try:
        enums = UnitTestTypes.get_unittest_enum_class()
    except:
        pass

    enums = enums if enums else [DefaultUnitTestTypes]

    # default tests
    for enum in enums:
        for test_case in enum:
            test_cls = test_case.import_class()
            suite.addTest(test_cls('test'))

    return suite



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: November 9th 2020

Modified By: hsky77
Last Updated: November 21st 2020 18:05:42 pm
'''

from .command import CommandProcessor

CommandProcessor().process_command()



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: November 21st 2020

Modified By: hsky77
Last Updated: November 21st 2020 21:36:16 pm
'''

from .app_mixin import WebApplicationMinin
from .config import ConfigControllerValidator
from .controller import ControllerType



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: September 4th 2020 14:12:47 pm
'''

import time
import random
import asyncio

from ..util import Worker, FunctionLoopWorker, FunctionQueueWorker
from .base import UnitTestCase


class WorkerTestCase(UnitTestCase):
    def test(self):
        self.test_pools()
        self.test_workers()

    def test_pools(self):
        """test worker pool both sync and async function"""
        from ..util import ExecutorFactory
        loop = asyncio.get_event_loop()
        worker_count = 1
        ap = ExecutorFactory(worker_limit=worker_count)

        def foo(index, count):
            self.assertIsNotNone(index)
            return index

        def foo_raise_exception(index, kwindex):
            self.assertIsNotNone(index)
            raise Exception('This is from foo_raise_exception()')

        result = ap.run_method(foo, 1, 1)
        self.assertEqual(result, 1)

        with ap.get_executor() as executor:
            result = executor.run_method(foo, 2, 1)
            self.assertEqual(result, 2)

            try:
                executor.run_method(foo_raise_exception, 1, 3)
            except Exception as e:
                self.assertEqual(str(e), 'This is from foo_raise_exception()')

        async def async_foo(index):
            result = await ap.run_method_async(foo, index, 1)
            self.assertEqual(result, index)

            try:
                await ap.run_method_async(foo_raise_exception, index, 3)
            except Exception as e:
                self.assertEqual(
                    str(e), 'This is from foo_raise_exception()')

            async with ap.get_executor() as executor:
                result = await executor.run_method_async(foo, index, 1)
                self.assertEqual(result, index)

                try:
                    await executor.run_method_async(foo_raise_exception, index, 3)
                except Exception as e:
                    self.assertEqual(
                        str(e), 'This is from foo_raise_exception()')

        futures = []
        count = 10000
        for i in range(count):
            futures.append(asyncio.ensure_future(async_foo(i)))

        loop.run_until_complete(asyncio.wait(futures))

        ap.dispose()
        self.assertLessEqual(ap.worker_count, worker_count)

    def test_workers(self):
        """test function and callback have been executed properly, it should takes 2~3 secs"""

        self.check_exp = False
        self.loop_count = 0

        def foo(index, kwindex):
            self.assertIs(index, 1)
            self.assertIs(kwindex, 2)
            self.loop_count = self.loop_count + 1
            return index

        def foo_raise_exception(index, kwindex):
            self.assertIs(index, 1)
            self.assertIs(kwindex, 2)
            self.loop_count = self.loop_count + 1
            raise Exception('This is from foo_raise_exception()')

        def on_finish(result):
            self.assertIs(result, 1)

        def on_exception(e):
            self.assertEqual(str(e), 'This is from foo_raise_exception()')
            self.check_exp = True

        with Worker() as worker:
            worker.run_method(foo, 1, kwindex=2,
                              on_finish=on_finish, on_exception=on_exception)

            while worker.is_func_running:
                pass

            self.check_exp = False
            worker.run_method(foo_raise_exception, 1, kwindex=2,
                              on_finish=on_finish, on_exception=on_exception)

            while worker.is_func_running:
                pass

            self.assertTrue(self.check_exp)

        with FunctionQueueWorker() as worker:
            self.check_exp = False
            worker.run_method(foo, 1, kwindex=2, on_finish=on_finish,
                              on_exception=on_exception)

            worker.run_method(foo_raise_exception, 1, kwindex=2, on_finish=on_finish,
                              on_exception=on_exception)

            start_time = time.time()
            while worker.pending_count > 0:
                self.assertGreaterEqual(3, time.time() - start_time)

            self.assertTrue(self.check_exp)

        with FunctionLoopWorker(loop_interval_seconds=0.1) as worker:
            self.check_exp = False
            self.loop_count = 0
            wait_time = 0.5

            worker.run_method(foo, 1, kwindex=2, on_method_runned=on_finish,
                              on_exception=on_exception)

            start_time = time.time()
            while time.time() - start_time < wait_time:
                pass

            self.assertGreaterEqual(self.loop_count, 2)

            self.loop_count = 0
            worker.run_method(foo_raise_exception, 1, kwindex=2, on_method_runned=on_finish,
                              on_exception=on_exception)

            start_time = time.time()
            while time.time() - start_time < wait_time:
                pass

            self.assertGreaterEqual(self.loop_count, 2)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: May 7th 2022 09:24:22 am
'''


import os

from enum import Enum
from typing import List, Dict, Tuple

from .utils import join_path
from .constants import LocalCode_No_Code, LocalCode_Duplicated_Code, LocalCode_Local_Pack_Parsing_Error, LocalCode_Message_Format_Invalid


class Localization():
    """convert message to localized language message"""
    default_code = str(LocalCode_No_Code)

    def __init__(self, lang: str = 'en'):
        self.__mapping = {}
        self.__csvs = []
        self.set_language(lang)
        self.__languages = set()

    def set_language(self, lang: str) -> None:
        self.__lang = lang

    @property
    def current_language(self) -> str:
        return self.__lang

    def get_info(self) -> Dict:
        """return dict shows how many language avaliable and the codes loaded"""
        return {
            'current_language': self.current_language,
            'loaded_languages': list(self.__languages),
            'codes': len(self.__mapping),
            'files_loaded': self.__csvs
        }

    def import_csvs_from_directory(self, dir: str, encoding: str = 'utf-8') -> None:
        """import coded message from all csv files of indicated directory"""
        self.import_csv([join_path(dir, f) for f in os.listdir(
            dir) if '.csv' in f and os.path.isfile(join_path(dir, f))])

    def import_csv(self, files: List[str], encoding: str = 'utf-8', replace_duplicated_code: bool = True) -> None:
        """import coded message from csv file"""
        import re
        for path in files:
            with open(path, 'r', encoding=encoding) as f:
                self.__csvs.append(path)
                lines = f.readlines()

                if len(lines) > 0:
                    langs = [x.replace('\n', '') for x in lines.pop(
                        0).split(',') if not 'code' in x]

                for line in lines:
                    line = [x for x in re.split(',"(.*?)"|,', line) if
                            x is not None and not x == '' and not x == '\n']
                    if not len(line) - 1 == len(langs):
                        raise SyntaxError(self.get_message(
                            LocalCode_Local_Pack_Parsing_Error, path))
                    idx = 1
                    for lang in langs:
                        self.__languages.add(lang)
                        if not line[0] in self.__mapping:
                            self.__mapping[line[0]] = {}

                        if not lang in self.__mapping[line[0]]:
                            self.__mapping[line[0]][lang] = line[idx].replace(
                                '\n', '')
                        else:
                            if replace_duplicated_code:
                                self.__mapping[line[0]][lang] = line[idx].replace(
                                    '\n', '')
                            else:
                                raise KeyError(self.get_message(
                                    LocalCode_Duplicated_Code, line[0], lang))

                        idx = idx + 1

    def has_message(self, code: str) -> bool:
        code = str(code)
        return code in self.__mapping

    def get_message(self, code: str, *strings) -> str:
        """convert to localized message via code and following parameters"""
        code = str(code)
        if not self.__lang in self.__languages:
            raise KeyError(
                'language: {}, code: {} does not exist'.format(self.__lang, code))

        if code in self.__mapping:
            lang = self.__lang
            if not self.__lang in self.__mapping[code]:
                lang = 'en'

            try:
                return self.__mapping[code][lang].format(*strings)
            except:
                return self.__mapping[str(LocalCode_Message_Format_Invalid)][lang].format(code, strings)
        else:
            lang = self.__lang
            if not self.__lang in self.__mapping[self.default_code]:
                lang = 'en'

            return self.__mapping[self.default_code][lang].format(self.__lang, code)


BaseLocal = Localization()



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: September 4th 2020 14:12:34 pm
'''

import asyncio
from enum import Enum

from ..util import Callbacks
from .base import UnitTestCase


class TestCallbackType(Enum):
    Event_A = 'a'
    Event_A_Async = 'a_async'


class UtilTestCase(UnitTestCase):
    def test(self):
        self.test_callback()

    def test_callback(self, callbacks: Callbacks = None):
        cb = callbacks or Callbacks(TestCallbackType)
        cb.add_callback(TestCallbackType.Event_A, UtilTestCase.test_func)
        cb.add_callback(TestCallbackType.Event_A_Async,
                        UtilTestCase.test_func_async)

        cb.execute_callback(TestCallbackType.Event_A, self, 1, kwindex=2)

        loop = asyncio.get_event_loop()
        loop.run_until_complete(cb.execute_callback_async(
            TestCallbackType.Event_A_Async, self, 1, kwindex=2))

    def test_func(self, i, kwindex):
        self.assertEqual(i, 1)
        self.assertEqual(kwindex, 2)

    async def test_func_async(self, i, kwindex):
        self.assertEqual(i, 1)
        self.assertEqual(kwindex, 2)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: November 21st 2020 20:52:41 pm
'''

# reserved folder names
Component_Module_Folder = 'component'
Controller_Module_Folder = 'controller'
Unittest_Module_Folder = 'unit_test'
Dependency_Folder = 'dependency'

# reserved file names
Localization_File = 'local.csv'
Project_Config_File = 'project_config.yml'
Project_Pack_File = 'pack.yml'
Project_Requirement_File = 'requirements.txt'

# project
LocalCode_Not_Subclass = 100
LocalCode_File_Not_Found = 101

# project config
LocalCode_Parameter_Required = 110
LocalCode_Parameter_Type_Error = 111
LocalCode_Setup_Error = 112
LocalCode_Invalid_Parameter = 113
LocalCode_Copied_Name_Equal_To_Origin = 114

# components
LocalCode_Component_Duplicated_Key = 120
LocalCode_Failed_To_Load_Component = 121
LocalCode_Component_Type_Not_Exist = 122

# controllers' localization code
LocalCode_Failed_To_Load_Controller = 130

# controller
LocalCode_Missing_Required_Parameter = 140
LocalCode_Upload_Success = 141

LocalCode_Connect_Failed = 150



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: October 15th 2021 13:47:00 pm
'''

import logging
from threading import Lock

LOG_FORMAT = '%(asctime)s %(levelname)-7s %(name)s - %(message)s'


class BaseSyncLogger(logging.Logger):
    def __init__(self, name: str, level: int = logging.INFO):
        super().__init__(name, level)
        self.lock = Lock()

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False):
        with self.lock:
            super()._log(level, msg, args, exc_info=exc_info, extra=extra, stack_info=stack_info)


logging.setLoggerClass(BaseSyncLogger)
logging.basicConfig(format=LOG_FORMAT)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

This module defines the base component classes

Modified By: hsky77
Last Updated: May 7th 2022 14:50:46 pm
'''


from audioop import reverse
from typing import Dict, Tuple, Any, Union, List
from inspect import iscoroutinefunction

from ...util import DynamicTypeEnum, BaseLocal
from ..constants import LocalCode_Not_Subclass, LocalCode_Component_Type_Not_Exist, Component_Module_Folder


class ComponentTypes(DynamicTypeEnum):
    """base component enum class"""

    @staticmethod
    def get_component_enum_class() -> List[DynamicTypeEnum]:
        try:
            return DynamicTypeEnum.get_dynamic_class_enum_class(Component_Module_Folder)
        except:
            pass

    def import_class(self):
        return super().import_class(cls_type=Component)


class Component():
    """base component class"""

    def __init__(self, component_type: ComponentTypes):
        self.component_type = component_type
        self.component_key = component_type.enum_key

    def init(self, component_manager, *arugs, **kwargs) -> None:
        """called when component_manager create component objects"""
        raise NotImplementedError()

    def info(self) -> Dict:
        """define what meta information of component should be return"""
        return {'component': type(self).__name__, 'info': None}

    def dispose(self, component_manager) -> None:
        pass


class ComponentManager():
    """class to store and manage components"""

    def __init__(self):
        self.__components = {}

    @property
    def components(self) -> List[Component]:
        return [v for k, v in self.__components.items()]

    @property
    def info(self) -> Dict:
        info = {}
        for component in self.components:
            info[component.component_type.enum_key] = component.info()
        return info

    async def dispose_components(self) -> None:
        return await self.boardcast_async('dispose', self)

    def boardcast(self, method: str, *arugs, **kwargs) -> List[Tuple[ComponentTypes, Any]]:
        """
        this function invokes the non-awaitable methods of stored components and

        return a list of returns from each component methods
        """
        result = []
        for component in self.components:
            if hasattr(component, method):
                func = getattr(component, method)
                if callable(func):
                    if not iscoroutinefunction(func):
                        result.append(
                            (component.component_type, func(*arugs, **kwargs)))
        return result

    async def boardcast_async(self, method: str, *arugs, **kwargs) -> List[Tuple[ComponentTypes, Any]]:
        """
        this function invokes both awaitable and non-awaitable methods of stored components and 

        return a list of returns from each component
        """
        result = []
        for component in self.components:
            if hasattr(component, method):
                func = getattr(component, method)
                if callable(func):
                    if iscoroutinefunction(func):
                        result.append((component.component_type, await func(*arugs, **kwargs)))
                    else:
                        result.append(
                            (component.component_type, func(*arugs, **kwargs)))
        return result

    def invoke(self, enum_type: Union[str, ComponentTypes], method: str, *arugs, **kwargs) -> Any:
        """execute component mehtod by giving the method name and arguments"""
        component = self.get_component(enum_type)
        func = getattr(component, method, None)
        if callable(func):
            if not iscoroutinefunction(func):
                return func(*arugs, **kwargs)

    async def invoke_async(self, enum_type: Union[str, ComponentTypes], method: str, *arugs, **kwargs) -> Any:
        """asynchronously execute component mehtod by giving the method name and arguments"""
        component = self.get_component(enum_type)
        func = getattr(component, method, None)
        if callable(func):
            if iscoroutinefunction(func):
                return await func(*arugs, **kwargs)

    def set_component(self, component: Component) -> None:
        """add or replace component object"""
        if isinstance(component, Component):
            self.__components[component.component_key] = component
        else:
            raise TypeError(BaseLocal.get_message(
                LocalCode_Not_Subclass, type(component), Component))

    def get_component(self, enum_type: Union[str, ComponentTypes]) -> Component:
        """return stored component object"""
        key = self.__get_component_key(enum_type)
        if key in self.__components:
            return self.__components[key]

        raise KeyError(BaseLocal.get_message(
            LocalCode_Component_Type_Not_Exist, enum_type))

    def has_component(self, enum_type: Union[str, ComponentTypes]) -> bool:
        return self.__get_component_key(enum_type) in self.__components

    def sort_components(self, order_list: List[ComponentTypes]):
        """
        sort component object by "order_list". default sorting with ComponentTypes order

        notes: considering the component dependencies, use ComponentTypes enum class to make the order to dispose components when server shut down
        """
        orders = []
        for x in order_list:
            reverse_array = [k for k in reversed([y.enum_key for y in x])]
            if reverse_array is not None:
                orders = orders + reverse_array
        self.__components = {x: self.__components[x] for x in orders if x in self.__components}

    def __get_component_key(self, enum_type: Union[str, ComponentTypes]) -> str:
        if isinstance(enum_type, ComponentTypes):
            return enum_type.enum_key
        return enum_type



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: October 27th 2021 14:42:57 pm
'''

import os
import shutil
import setuptools
from hyssop import Version, __name__

if os.path.isdir('dist'):
    shutil.rmtree('dist')

if os.path.isdir('hyssop.egg-info'):
    shutil.rmtree('hyssop.egg-info')

if os.path.isdir('build'):
    shutil.rmtree('build')

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=__name__,
    version=Version,
    author="hsky77",
    author_email="howardlkung@gmail.com",
    description="component-based project hierarchy and utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hsky77/hyssop",
    packages=setuptools.find_packages(include=('hyssop', 'hyssop.*')),
    license="MIT License",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['PyYAML>=5.1.1'],
    python_requires='>=3.6',
    package_data={'': ['*.yml', '*.yaml', '*.csv']}
)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

This module defines the classes inherit threading.Thread to execute functions:

    - Worker: Executes function once
    - FunctionLoopWorker: Loops a single function until stop() is called
    - FunctionQueueWorker: Executes functions in queue

Modified By: hsky77
Last Updated: August 27th 2020 13:04:38 pm
'''

import time
from typing import Callable, Any
from threading import Thread, Condition, Lock


class _BaseWorker(Thread):
    """hyssop-customized python threading.Thread class"""

    def __init__(self, name: str = None):
        super().__init__(name=name)
        self._running = False
        self._disposed = False

        self._run_method_lock = Lock()
        self.__resource_lock = None

        self._paused = False
        self._pause_lock = Lock()
        self._pause_cond = Condition(Lock())
        self.pause()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dispose()

    @property
    def is_started(self):
        return self._started.is_set()

    @property
    def resource_lock(self) -> Lock:
        """use this lock access resource synchronously"""
        self.__resource_lock = self.__resource_lock if self.__resource_lock else Lock()
        return self.__resource_lock

    def dispose(self) -> None:
        if not self._disposed:
            self._disposed = True
            self._running = False
            self.resume()

    def start(self) -> None:
        self._running = True
        super().start()

    def pause(self) -> None:
        if not self._paused:
            self._paused = True
            with self._pause_lock:
                self._pause_cond.acquire()

    def resume(self) -> None:
        if self._paused:
            self._paused = False
            with self._pause_lock:
                self._pause_cond.notify()
                self._pause_cond.release()

    def run(self) -> None:
        """do not directly call this function, it will be called by the thread after thread starts"""
        while self._running:
            with self._pause_cond:
                while self._paused:
                    self._pause_cond.wait()
            try:
                self._run()
            finally:
                self.pause()

    def _run(self) -> Any:
        """
        override this function to run something in thread, handle exceptions here to keep the thread alive
        """
        NotImplementedError()

    def _execute_function(self, func, *args, **kwargs) -> Any:
        """define how to execute the function"""
        return func(*args, **kwargs)


class Worker(_BaseWorker):
    """hyssop-customized python threading.Thread class to execute functions"""

    def __init__(self, name: str = None):
        super().__init__(name=name)
        self._func = None
        self._args = None
        self._kwargs = None
        self._on_finish = None
        self._on_exception = None

    def __enter__(self):
        return self

    @property
    def is_func_running(self) -> bool:
        return self._func is not None

    def run_method(self,
                   func: Callable,
                   *args,
                   on_finish: Callable[[Any], None] = None,
                   on_exception: Callable[[Exception], None] = None,
                   **kwargs) -> bool:
        """return True if function will be executed, False if there is a function is running"""
        if self._disposed:
            return False

        result = False
        with self._run_method_lock:
            if callable(func) and not self.is_func_running:
                self._func = func
                self._args = args
                self._kwargs = kwargs
                self._on_finish = on_finish
                self._on_exception = on_exception

                if not self.is_started:
                    self.start()

                self.resume()
                result = True
        return result

    def _run(self) -> Any:
        if callable(self._func):
            try:
                self._result = self._execute_function(
                    self._func, *self._args, **self._kwargs)
                if callable(self._on_finish):
                    self._on_finish(self._result)
            except Exception as e:  # handle exception here to keep the thread alive
                if callable(self._on_exception):
                    self._on_exception(e)
            finally:
                self._func = None
                self._args = None
                self._kwargs = None
                self._on_finish = None
                self._on_exception = None


class FunctionQueueWorker(_BaseWorker):
    """worker that queues functions to execute"""

    def __init__(self, name: str = None):
        super().__init__(name)
        self.__tasks = []

    def __enter__(self):
        return self

    @property
    def pending_count(self) -> int:
        return len(self.__tasks)

    def run_method(self,
                   func: Callable,
                   *args,
                   on_finish: Callable[[Any], None] = None,
                   on_exception: Callable[[Exception], None] = None,
                   **kwargs) -> None:
        """func will be queued and run when the worker thread is free"""
        if self._disposed:
            return None

        if callable(func):
            self.__tasks.append({
                'func': func,
                'on_finish': on_finish,
                'on_exception': on_exception,
                'args': args,
                'kwargs': kwargs
            })

            if not self.is_started:
                self.start()

            self.resume()

    def _run(self) -> Any:
        while len(self.__tasks) > 0:
            task = self.__tasks.pop(0)

            try:
                result = self._execute_function(
                    task['func'], *task['args'], **task['kwargs'])
                if callable(task['on_finish']):
                    task['on_finish'](result)
            except Exception as e:
                if callable(task['on_exception']):
                    task['on_exception'](e)


class FunctionLoopWorker(_BaseWorker):
    """worker class loops a single function in one period before calling stop()"""

    def __init__(self, name: str = None, loop_interval_seconds: float = 1.0):
        super().__init__(name=name)
        self.__loop_interval_seconds = loop_interval_seconds
        self.__on_method_runned = None
        self.__on_exception = None
        self.stop()

    def __enter__(self):
        return self

    def dispose(self):
        super().dispose()
        self.stop()
        self.join()

    def stop(self):
        """stop looping"""
        self._func = None
        self._args = None
        self._kwargs = None
        self.__on_method_runned = None
        self.__on_exception = None

    def run_method(self,
                   func: Callable,
                   *args,
                   on_method_runned: Callable[[Any], None] = None,
                   on_exception: Callable[[Exception], None] = None,
                   **kwargs) -> None:
        """start looping function"""
        if self._disposed:
            return None

        with self._run_method_lock:
            if self._func is None:
                self._func = func
                self._args = args
                self._kwargs = kwargs
                self.__on_method_runned = on_method_runned
                self.__on_exception = on_exception

                if not self.is_started:
                    self._running = True
                    self.start()

                self.resume()

    def _run(self):
        while callable(self._func):
            try:
                result = self._execute_function(
                    self._func, *self._args, **self._kwargs)
                if callable(self.__on_method_runned):
                    self.__on_method_runned(result)
            except Exception as e:
                if callable(self.__on_exception):
                    self.__on_exception(e)
            finally:
                time.sleep(self.__loop_interval_seconds)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

This module contains the "yaml" configurable component classes for hyssop application.

    - to create Project components:

        0. the hierarchy of folders and files looks like:

            server_directory/
                server_config.yaml
                component/
                    __init__.py
                    foo.py

        1. use ComponentTypes enum class to define the extend components such as:

            class ComponentExtension(ComponentTypes):
                # tuple(<component_key>, <package_route>, <class_name_in_the_py_file>)

                Foo = ('foo', 'foo', 'Foo')

        2. in "foo.py" contains the class code:

            from hyssop.project.component import Component, ComponentManager
            from . import ComponentExtension

            class Foo(Component):
                def __init__(self):
                    super().__init__(ComponentExtension.Foo)

                def init(self, component_manager: ComponentManager, p1, *arugs, **kwargs) -> None:
                    self.p1 = p1

        3. setup component block of "server_config.yaml" to tell hyssop server load the extend components "Foo":

            component:              # block to setup component
                foo:                # component_key to load
                    p1: xxxx        # parameter p1 of Foo.init()

Modified By: hsky77
Last Updated: January 17th 2021 17:56:39 pm
'''

from typing import Dict, List, Union

from .base import ComponentTypes, Component, ComponentManager
from ...util import BaseLocal
from ..constants import LocalCode_Component_Duplicated_Key, LocalCode_Failed_To_Load_Component

from .config import ConfigComponentValidator


class DefaultComponentTypes(ComponentTypes):
    """server loads all components of this enum type when start"""

    Localization = ('localization', 'default', 'LocalizationComponent')
    Logger = ['logger', 'default', 'LoggerComponent']
    Callback = ('callback', 'default', 'CallbackComponent')
    Executor = ('executor', 'default', 'ExecutorComponent')


DefaultComponentEnums = [DefaultComponentTypes]


def add_default_component_types(component_type: ComponentTypes):
    DefaultComponentEnums.insert(0, component_type)


def __create_optional_components(component_manager: ComponentManager, component_settings: Dict, component_types: ComponentTypes, project_dir: str) -> None:
    for key in component_settings:
        for component_type in component_types:
            comp_type = None
            try:
                comp_type = component_type(key)
            except:
                continue
            if comp_type is not None:
                comp = comp_type.import_class()(comp_type)
                component_manager.set_component(comp)
                break

    for key in component_settings:
        comp_type = None
        for component_type in component_types:
            try:
                comp_type = component_type(key)
                break
            except:
                continue

        if comp_type:
            component_manager.invoke(comp_type, 'init',
                                     component_manager,
                                     **(component_settings[comp_type.enum_key] or {}),
                                     project_dir=project_dir)


def create_component_manager(project_dir: str, component_settings: Union[Dict, None] = None) -> ComponentManager:
    component_manager = ComponentManager()

    # extensions
    ext_comp_types = ComponentTypes.get_component_enum_class()

    # default components
    for default_enums in DefaultComponentEnums:
        for default_type in default_enums:
            comp = default_type.import_class()(default_type)
            component_manager.set_component(comp)

    # init
    for default_enums in DefaultComponentEnums:
        for default_type in default_enums:
            if component_settings and default_type.enum_key in component_settings:
                component_manager.invoke(default_type, 'init',
                                         component_manager,
                                         **(component_settings[default_type.enum_key] or {}),
                                         project_dir=project_dir)
            else:
                component_manager.invoke(
                    default_type, 'init', component_manager, project_dir=project_dir)

    sort_types = DefaultComponentEnums

    # validate config
    ConfigComponentValidator(component_settings)

    if ext_comp_types is not None:
        sort_types = ext_comp_types + sort_types

        # check duplicated key:
        for r_type in sort_types:
            for key in r_type:
                for l_type in sort_types:
                    d_key = None
                    if r_type is not l_type:
                        try:
                            d_key = l_type(key.enum_key)
                        except:
                            continue
                    if d_key is not None:
                        raise KeyError(BaseLocal.get_message(
                            LocalCode_Component_Duplicated_Key, l_type, r_type, key.enum_key))

        if component_settings:
            __create_optional_components(
                component_manager, component_settings, ext_comp_types, project_dir)

    # check componet load failed
    if component_settings:
        for key in component_settings:
            checked = False
            for component in component_manager.components:
                if key == component.component_type.enum_key:
                    checked = True
            if not checked:
                raise ImportError(BaseLocal.get_message(
                    LocalCode_Failed_To_Load_Component, project_dir, key))

    # sort with enums order
    component_manager.sort_components(sort_types)

    return component_manager


def add_module_default_logger(logger_names: List[str]) -> None:
    from .default import LoggerComponent
    LoggerComponent.default_loggers = LoggerComponent.default_loggers + logger_names



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: October 25th 2021 09:14:48 am
'''

import os
import sys
import argparse
from typing import Callable

from . import Module_Path, Version
from .util import get_class, join_path, join_to_abs_path
from .project.constants import (
    Unittest_Module_Folder, Project_Config_File, Project_Pack_File, Project_Requirement_File,
    Component_Module_Folder, Controller_Module_Folder)


class CommandProcessor:
    Command_Test_Project = 'test'
    Command_Create_Project = 'create'
    Command_Show_Version = 'version'
    Command_Pack_Project = 'pack'

    def __init__(self):
        self.project_dir = None
        self.__create_command_parser()

    @property
    def args_key_project_directory(self) -> str:
        return 'project_directory'

    @property
    def project_component_dir(self) -> str:
        return join_path(self.project_dir, Component_Module_Folder)

    @property
    def project_controller_dir(self) -> str:
        return join_path(self.project_dir, Controller_Module_Folder)

    @property
    def project_unitetest_dir(self) -> str:
        return join_path(self.project_dir, Unittest_Module_Folder)

    @property
    def project_config_file(self) -> str:
        return join_path(self.project_dir, Project_Config_File)

    @property
    def project_pack_file(self) -> str:
        return join_path(self.project_dir, Project_Pack_File)

    @property
    def project_requirement_file(self) -> str:
        return join_path(self.project_dir, Project_Requirement_File)

    def process_command(self):
        self.__parse_command()

        if hasattr(self.args, 'command'):
            if hasattr(self, self.args.command):
                func = getattr(self, self.args.command)
                if callable(func):
                    func()
                else:
                    self.parser.print_help()
        else:
            self.parser.print_help()

    def test(self):
        import unittest
        from .unit_test import get_test_suite
        runner = unittest.TextTestRunner()
        runner.run(get_test_suite())

    def pack(self) -> None:
        from .project.pack import HyssopPack
        if self.project_dir:
            HyssopPack().pack(self.project_dir, self.args.o,
                              prepare_wheels=self.args.add_wheels,
                              compile_py=not self.args.decompile_pyc)

    def version(self):
        print('hyssop {}'.format(Version))

    def create(self):
        self.project_dir = self.project_dir if self.project_dir else 'hello_world'

        if not os.path.isdir(self.project_dir):
            os.makedirs(self.project_dir)

        self._create_project_component_files()
        self._create_project_controller_files()
        self._create_project_config_files()
        self._create_project_test_files()
        self._create_project_pack_files()
        self._create_project_requirement_files()

        print('project created at', os.path.abspath(self.project_dir))

    def _create_project_component_files(self):
        if not os.path.isdir(self.project_component_dir):
            os.makedirs(self.project_component_dir)

        with open(join_path(self.project_component_dir, '__init__.py'), 'w') as f:
            f.write('''\
from hyssop.project.component import ComponentTypes, ConfigComponentValidator

from hyssop.project.config_validator import ConfigContainerMeta, ConfigElementMeta

# add hello validator to component config validator
ConfigComponentValidator.set_cls_parameters(
    ConfigContainerMeta('hello', False,
        ConfigElementMeta('p1', str, True) # validate HelloComponent's 'p1' argument is required and string type
    )
)

class HelloComponentTypes(ComponentTypes):
    Hello = ('hello', 'hello', 'HelloComponent')
''')

        with open(join_path(self.project_component_dir, 'hello.py'), 'w') as f:
            f.write('''\
from hyssop.project.component import Component
from . import HelloComponentTypes

class HelloComponent(Component):
    def init(self, component_manager, p1, *arugs, **kwargs) -> None:
        print('init Hello component load from', __package__, 'and the parameters p1:', p1)

    def hello(self):
        return 'Hello World, This is hyssop generate hello component'
''')

    def _create_project_controller_files(self):
        pass

    def _create_project_test_files(self):
        if not os.path.isdir(self.project_unitetest_dir):
            os.makedirs(self.project_unitetest_dir)

        with open(join_path(self.project_unitetest_dir, '__init__.py'), 'w') as f:
            f.write('''\
from hyssop.unit_test import UnitTestTypes

class UTTypes(UnitTestTypes):
    UT1 = ('ut1', 'ut1', 'UT1TestCase')
''')

        with open(join_path(self.project_unitetest_dir, 'ut1.py'), 'w') as f:
            f.write('''\
from hyssop.unit_test import UnitTestCase

class UT1TestCase(UnitTestCase):
    def test(self):
        # implement unit test here...
        import os
        from component import HelloComponentTypes
        from hyssop.project.mixin import ProjectMixin

        project = ProjectMixin()
        project.load_project(os.path.dirname(os.path.dirname(__file__)))
        comp = project.component_manager.get_component(
            HelloComponentTypes.Hello)
        print(comp.hello())
''')

    def _create_project_config_files(self):
        with open(self.project_config_file, 'w') as f:
            f.write('''\
name: hyssop Project
debug: False
component:
  hello: 
    p1: 'This is p1'
''')

    def _create_project_pack_files(self):
        with open(self.project_pack_file, 'w') as f:
            f.write('''
# This is packing list indicated what are the files should be pack
# If this file does not exist under the project folder, all of the files under the folder will be packed

include:
# List absolute or relative path of additional file or directory to be packed
# - example.txt
# - example_dir

exclude:
# List absolute or relative path of file, directory, or file extension to be ignored.
- '.log'
''')

    def _create_project_requirement_files(self):
        # requirement
        from . import __name__, Version
        with open(self.project_requirement_file, 'w') as f:
            f.write('{}>={}'.format(__name__, Version))

    def __create_command_parser(self):
        self.parser = argparse.ArgumentParser(prog='hyssop')
        self.command_parsers = self.parser.add_subparsers(title='command')

        test_parser = self.command_parsers.add_parser(
            CommandProcessor.Command_Test_Project, help='test hyssop library or specfied project directory path')
        test_parser.add_argument(self.args_key_project_directory, nargs='?',
                                 help='project directory path')
        test_parser.set_defaults(command=CommandProcessor.Command_Test_Project)

        make_serv_parser = self.command_parsers.add_parser(
            CommandProcessor.Command_Create_Project, help='create a project template with specfied project directory path')
        make_serv_parser.add_argument(self.args_key_project_directory,
                                      help='project directory path')
        make_serv_parser.set_defaults(
            command=CommandProcessor.Command_Create_Project)

        pack_parser = self.command_parsers.add_parser(
            CommandProcessor.Command_Pack_Project, help='pack project with specfied project directory path')
        pack_parser.add_argument(self.args_key_project_directory,
                                 help='project directory path')
        pack_parser.add_argument(
            '-o', help='specify output compressed file path', default=None)
        pack_parser.add_argument('-w', '--add_wheels', action='store_true',
                                 help='add dependency wheel files')
        pack_parser.add_argument('-d', '--decompile_pyc', action='store_true',
                                 help='disable compile .py to .pyc')
        pack_parser.set_defaults(command=CommandProcessor.Command_Pack_Project)

        version_serv_parser = self.command_parsers.add_parser(
            CommandProcessor.Command_Show_Version, help='print version number to console')
        version_serv_parser.set_defaults(
            command=CommandProcessor.Command_Show_Version)

    def __parse_command(self):
        sys.path.insert(0, os.path.dirname(
            os.path.dirname(os.path.abspath(__file__))))
        self.args = self.parser.parse_args()

        if hasattr(self.args, self.args_key_project_directory):
            if self.args.project_directory:
                self.project_dir = join_to_abs_path(
                    self.args.project_directory)
                sys.path.insert(0, os.path.abspath(self.project_dir))
                sys.path.insert(0, os.path.dirname(
                    os.path.abspath(self.project_dir)))



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: November 21st 2020 21:18:14 pm
'''

import time
from unittest import TestCase
from typing import List

from ..util import DynamicTypeEnum

from ..project.constants import Component_Module_Folder, Controller_Module_Folder


class UnitTestTypes(DynamicTypeEnum):
    """base abstract unitest enum class"""

    @staticmethod
    def get_unittest_enum_class() -> List[DynamicTypeEnum]:
        from ..project.constants import Unittest_Module_Folder
        try:
            return DynamicTypeEnum.get_dynamic_class_enum_class(Unittest_Module_Folder)
        except:
            pass

    def import_class(self):
        return super().import_class(cls_type=UnitTestCase)


class UnitTestCase(TestCase):
    """hyssop unittest case abstract class"""

    def test(self):
        raise NotImplementedError()
